 // ------------------------------------------------------------------------
// Due CAN Layer - Created for easier interfacing
// ------------------------------------------------------------------------
#include <Arduino.h>
#include "mcp2515.h"

// CAN Ports Setup
// ------------------------------------------------------------------------
#define CAN_RX_INTERRUPT_MODE     1   // 1-ON; 0-OFF

// espBerry Settings (will need modification for other ESP32 systems)
// ------------------------------------------------------------------------
// Define CAN Ports and their Chip Select/Enable
MCP2515 CAN0(5);            // CAN0 interface CS
MCP2515 CAN1(17);           // CAN1 interface CS
// Interrupt Signals
const int CAN0_INT = 26;    // RPi Pin 16 - GPIO23 - IO26
const int CAN1_INT = 34;    // RPi Pin 22 - GPIO25 - IO34
// ------------------------------------------------------------------------
#if CAN_RX_INTERRUPT_MODE == 1

  // Receive Buffer Architecture
  #define RXBUFFERSIZE      64
  typedef struct CANRxBuffer
  {
      unsigned long lMsgID;
      byte cLen;
      byte cExt;
      byte pBuffer[8];
  };

  CANRxBuffer pCANRxBuffer_0[RXBUFFERSIZE];
  CANRxBuffer pCANRxBuffer_1[RXBUFFERSIZE];

  int nWrPointer_0 = 0;
  int nWrPointer_1 = 0;
  int nRdPointer_0 = 0;
  int nRdPointer_1 = 0;
// ------------------------------------------------------------------------

// ------------------------------------------------------------------------
// ISR CAN0
// ------------------------------------------------------------------------
void isrCAN0(void)
{
  // Declarations
  struct can_frame canMsg;

  // Read data: len = data length, buf = data byte(s)
  if(CAN0.readMessage(&canMsg) == CAN_OK)    
  {
    // Store data frame into Rx buffer
    pCANRxBuffer_0[nWrPointer_0].lMsgID = canMsg.can_id & 0x1FFFFFFF;;
    pCANRxBuffer_0[nWrPointer_0].cExt 
                = ((canMsg.can_id & 0x80000000) >> 31);
    pCANRxBuffer_0[nWrPointer_0].cLen = canMsg.can_dlc;

    for(int nIndex = 0; nIndex < (int)canMsg.can_dlc; nIndex++)
        pCANRxBuffer_0[nWrPointer_0].pBuffer[nIndex] = canMsg.data[nIndex];

    // Check the write pointer
    if(nWrPointer_0 == RXBUFFERSIZE-1)
        nWrPointer_0 = 0;
    else
        nWrPointer_0++;

  }// end if

}// end isrCAN0

// ------------------------------------------------------------------------
// ISR CAN1
// ------------------------------------------------------------------------
void isrCAN1(void)
{
  // Declarations
  struct can_frame canMsg;

  // Read data: len = data length, buf = data byte(s)
  if(CAN1.readMessage(&canMsg) == CAN_OK)    
  {
    // Store data frame into Rx buffer
    pCANRxBuffer_1[nWrPointer_1].lMsgID = canMsg.can_id & 0x1FFFFFFF;;
    pCANRxBuffer_1[nWrPointer_1].cExt = ((canMsg.can_id & 0x80000000) >> 31);
    pCANRxBuffer_1[nWrPointer_1].cLen = canMsg.can_dlc;

    for(int nIndex = 0; nIndex < (int)canMsg.can_dlc; nIndex++)
        pCANRxBuffer_1[nWrPointer_1].pBuffer[nIndex] = canMsg.data[nIndex];

    // Check the write pointer
    if(nWrPointer_1 == RXBUFFERSIZE-1)
        nWrPointer_1 = 0;
    else
        nWrPointer_1++;

  }// end if

}// end isrCAN1

#endif

// ------------------------------------------------------------------------
// Initialize CAN Controller
// ------------------------------------------------------------------------
// canSpeed = CAN_200KBPS/CAN_250KBPS... as defined in mcp2515.h
// canOpMode = CAN_OP_NORMAL/CAN_OP_SLEEP/CAN_OP_LOOPBACK/CAN_OP_LISTENONLY
// Returns ERROR_OK/ERROR_FAILINIT
//
byte canInit(const int canPort, const CAN_SPEED canSpeed, const int canOpMode)
{
  // Declarations
  byte cRetCode = CAN_FAILINIT;

  // ------------------------------------------------------------------------
  // CAN PORT 0
  // ------------------------------------------------------------------------
  if(canPort == 0)
  {

#if CAN_RX_INTERRUPT_MODE == 1

    // Reset read-write pointers
    nWrPointer_0 = 0;
    nRdPointer_0 = 0;

#endif

    // Activate Interrupt Digital Inputs
    pinMode(CAN0_INT, INPUT_PULLUP);

#if CAN_RX_INTERRUPT_MODE == 1

    // Set the ISRs for both CAN ports
    attachInterrupt(CAN0_INT, isrCAN0, FALLING);

#endif

    // Initialize MCP2515 running at 16MHz with a baudrate = canSpeed 
    // and the masks and filters disabled.
    // Available choices: MCP_8MHZ, MCP_16MHZ, MCP_20MHZ (mcp2515.h)
    CAN0.reset();
    if(CAN0.setBitrate(canSpeed, MCP_16MHZ) == CAN_OK)
    {
        switch(canOpMode)
        {
            case CAN_OP_NORMAL:
                cRetCode = CAN0.setNormalMode();
                break;
                
            case CAN_OP_SLEEP: 
                cRetCode = CAN0.setSleepMode(); 
                break;
                
            case CAN_OP_LOOPBACK:
                cRetCode = CAN0.setLoopbackMode();
                break;
                
            case CAN_OP_LISTENONLY:
                cRetCode = CAN0.setListenOnlyMode();
                break;

        }// end switch

        // Check for errors
        if(cRetCode != CAN_OK)
            cRetCode = CAN_FAILINIT;

    }// end if

  }// end if

  // ------------------------------------------------------------------------
  // CAN PORT 1
  // ------------------------------------------------------------------------
  if(canPort == 1)
  {

#if CAN_RX_INTERRUPT_MODE == 1

    // Reset read-write pointers
    nWrPointer_1 = 0;
    nRdPointer_1 = 0;

#endif

    // Activate Interrupt Digital Inputs
    pinMode(CAN1_INT, INPUT_PULLUP);

#if CAN_RX_INTERRUPT_MODE == 1

    // Set the ISRs for both CAN ports
    attachInterrupt(CAN1_INT, isrCAN1, FALLING);

#endif

    // Initialize MCP2515 running at 16MHz with a baudrate = canSpeed 
    // and the masks and filters disabled.
    CAN1.reset();
    if(CAN1.setBitrate(canSpeed) == CAN_OK)
    {
        switch(canOpMode)
        {
            case CAN_OP_NORMAL:
                cRetCode = CAN1.setNormalMode();
                break;
                
            case CAN_OP_SLEEP: 
                cRetCode = CAN1.setSleepMode(); 
                break;
                
            case CAN_OP_LOOPBACK:
                cRetCode = CAN1.setLoopbackMode();
                break;
                
            case CAN_OP_LISTENONLY:
                cRetCode = CAN1.setListenOnlyMode();
                break;

        }// end switch

        // Check for errors
        if(cRetCode != CAN_OK)
            cRetCode = CAN_FAILINIT;

    }// end if

  }// end if

  return cRetCode;

}// end canInit

// ------------------------------------------------------------------------
// Receive CAN message frame
// ------------------------------------------------------------------------
// cPort = 0/1 - other numbers will be ignored
//
// Returns CAN_OK/CAN_NOMSG
//
// Controller Area Network Identifier structure
// bit 0-28 : CAN identifier (11/29 bit)
// bit 29   : error message frame flag (0 = data frame, 1 = error message)
// bit 30   : remote transmission request flag (1 = rtr frame)
// bit 31   : frame format flag (0 = standard 11 bit, 1 = extended 29 bit)
//
byte canRx(byte cPort, unsigned long* lMsgID, byte* cMessageIDFormat, 
           byte* cData, byte* cDataLen)
{
  // Declarations
  byte cRetCode = CAN_NOMSG;
  struct can_frame canMsg;

#if CAN_RX_INTERRUPT_MODE == 0

  // Check for dataframe at CAN0
  if(cPort == 0)
  {
    // Read the message
    if((cRetCode = CAN0.readMessage(&canMsg)) == CAN_OK)
    {
      // Copy Message ID
      *lMsgID = canMsg.can_id & 0x1FFFFFFF;

      // Determine Message Format
      *cMessageIDFormat = ((canMsg.can_id & 0x80000000) >> 31);

      // Copy Data
      *cDataLen = canMsg.can_dlc;
      for(int nIndex = 0; nIndex < canMsg.can_dlc; nIndex++)
          cData[nIndex] = canMsg.data[nIndex];

    }// end if

  }// end if

  // Check for dataframe at CAN1
  if(cPort == 1)
  {
    // Read the message
    if((cRetCode = CAN1.readMessage(&canMsg)) == CAN_OK)
    {
      // Copy Message ID
      *lMsgID = canMsg.can_id & 0x1FFFFFFF;

      // Determine Message Format
      *cMessageIDFormat = ((canMsg.can_id & 0x80000000) >> 31);

      // Copy Data
      *cDataLen = canMsg.can_dlc;
      for(int nIndex = 0; nIndex < canMsg.can_dlc; nIndex++)
          cData[nIndex] = canMsg.data[nIndex];

    }// end if

  }// end if

#endif

#if CAN_RX_INTERRUPT_MODE == 1

  // Determine the port
  if(cPort == 0)
  {
    // Check for received CAN data frames
    if(nRdPointer_0 != nWrPointer_0)
    {
      // Data frame received; copy it
      *lMsgID = pCANRxBuffer_0[nRdPointer_0].lMsgID;
      *cMessageIDFormat = pCANRxBuffer_0[nRdPointer_0].cExt;
      *cDataLen = pCANRxBuffer_0[nRdPointer_0].cLen;

      for(int nIndex = 0; nIndex < (int)*cDataLen; nIndex++)
        cData[nIndex] = pCANRxBuffer_0[nRdPointer_0].pBuffer[nIndex];

      // Adjust the read pointer
      if(nRdPointer_0 == RXBUFFERSIZE-1)
        nRdPointer_0 = 0;
      else
        nRdPointer_0++;

      // Set the return code
      cRetCode = CAN_OK;

    }// end if

  }// end if

  if(cPort == 1)
  {
    // Check for received CAN data frames
    if(nRdPointer_1 != nWrPointer_1)
    {
      // Data frame received; copy it
      *lMsgID = pCANRxBuffer_1[nRdPointer_1].lMsgID;
      *cMessageIDFormat = pCANRxBuffer_1[nRdPointer_1].cExt;
      *cDataLen = pCANRxBuffer_1[nRdPointer_1].cLen;

      for(int nIndex = 0; nIndex < (int)*cDataLen; nIndex++)
        cData[nIndex] = pCANRxBuffer_1[nRdPointer_1].pBuffer[nIndex];

      // Adjust the read pointer
      if(nRdPointer_1 == RXBUFFERSIZE-1)
        nRdPointer_1 = 0;
      else
        nRdPointer_1++;

      // Set the return code
      cRetCode = CAN_OK;

    }// end if

  }// end if

  #endif

  return cRetCode;

}// end canRx

// ------------------------------------------------------------------------
// Transmit CAN message frame
// ------------------------------------------------------------------------
// cPort = 0/1 - other numbers will be ignored
// cExtended = CAN_STDID/CAN_EXTID
//
// Returns CAN_OK when successful, otherwise CAN_FAILTX as
// defined in mcp_can_dfs.h
//
byte canTx(byte cPort, unsigned long lMsgID, byte cExtendedFormat, byte* cData, 
           byte cDataLen)
{
  // Declarations
  byte cRetCode = CAN_FAILTX;
  struct can_frame canMsg;  

  // Copy the message ID and message format
  canMsg.can_id = ((long)cExtendedFormat << 31) + lMsgID;

  // Copy data length
  canMsg.can_dlc = cDataLen;

  // Copy the data
  for(int nIndex = 0; nIndex < cDataLen; nIndex++)
    canMsg.data[nIndex] = cData[nIndex];

  // Transmit the data frame
  if(cPort == 0)
    cRetCode = CAN0.sendMessage(&canMsg);

  if(cPort == 1)
    cRetCode = CAN1.sendMessage(&canMsg);

  return cRetCode;

}// end canTx

// End of File ------------------------------------------------------------
// ------------------------------------------------------------------------

